from typing import Any, List, Tuple, Type, TypeVar, Union, Dict
from numpy import ndarray
from pathlib import Path
from torch import Tensor
from ray import tune
import numpy as np
import torch
import yaml

# export list
__all__ = [
    # types
    "OneOrPair",
    "TensorOrArray",
    # generic utilities
    "DotDict",
    "Config",
    "get_np_or_torch",
    "pack_many",
    "repeat_test",
    "execute_with_probability",
    # tensor/array operations
    "to_numpy",
    "split_complex",
    "pack_complex",
]

# = = = = types

""" 
Generic variable
"""
T = TypeVar("T")

"""
Single or Pair of values of a certain type.
"""
OneOrPair = Union[T, Tuple[T, T]]

"""
Can be a torch Tensor or numpy ndarray.
"""
TensorOrArray = Union[Tensor, ndarray]

# = = = = generic utilities


class DotDict(dict):
    """
    A dict that allows dot notation
    for accessing to the elements
    """

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def copy(self) -> "DotDict":
        return DotDict(super().copy())


class Config:
    def __init__(self, config_path: Path) -> None:
        """
        A class to handle yaml configuration files.
        Inside a configuration many sections can be defined,
        with parameters within them.

        Parameters
        ----------
        config_path : Path
            Path to the yaml configuration file
        """
        # loading the configuration
        err_msg = "only YAML configurations are supported"
        assert config_path.suffix == ".yml", err_msg
        with open(config_path, "r") as f:
            self.config = yaml.safe_load(f)

    def get(
        self,
        section: str,
        parameter: str,
        _type: Type = str,
        default: Any = None,
    ) -> Any:
        """
        Gets a parameter from the configuration.

        Parameters
        ----------
        section : str
            Configuration section
        parameter : str
            Name of the parameter
        _type : Type, optional
            Type of the parameter, by default str
        default : Any, optional
            Default if the parameter does not exist, by default None

        Returns
        -------
        Fetched parameter
        """
        cfg = self.config

        try:
            # getting the section
            sec = cfg[section]
            # getting the parameter
            param = _type(sec[parameter])
        except KeyError:
            return default

        return param

    def get_ray_tune_params(self, section: str = "ray_tune") -> Dict:
        """
        Gets the ray tune parameters from a configuration.
        The parameters should be written in the section specified
        and be divided by sampling method.

        E.g.
        ray_tune:
            uniform:
                param_0: [0, 5]
            randn:
                param_1: [0, 1]


        Parameters
        ----------
        section : str, optional
            The parent section where the parameter space is defined,
            by default "ray_tune"

        Returns
        -------
        Dict
            Dictionary containing the ray tune parameters
        """
        try:
            cfg = self.config[section]
        except KeyError:
            # no ray tune section
            return {}

        sampling_methods = cfg.keys()
        params = {}

        # adding each hyperparameters parsing the
        # values passed to the corrispondent tune
        # sampling function f
        for sm in sampling_methods:
            f = getattr(tune, sm)
            _parsetype = lambda x: x if type(x) in (list, tuple) else float(x)
            params = params | {p: f(*map(_parsetype, v)) for p, v in cfg[sm].items()}
        return params


def get_np_or_torch(x: TensorOrArray):
    """
    Returns numpy or torch modules depending on the input

    Parameters
    ----------
    x : TensorOrArray
        Input

    Returns
    -------
    Module
        numpy or torch
    """
    if isinstance(x, Tensor):
        return torch
    else:
        torch.iscomplex = torch.is_complex
        torch.hanning = torch.hann_window
        np.cat = np.concatenate
        return np


def pack_many(*xss: List[List]) -> List[Tuple]:
    """
    Packs many Lists in one.

    Parameters
    ----------
    xss : List[List]

    Returns
    -------
    List[Tuple]
        Packed list
    """
    return list(zip(*xss))


# fmt: off
def repeat_test(times: int): 
    """
    Decorator to repeat a test n times.

    Parameters
    ----------
    times : int
        Number of repetitions
    """
    def repeatHelper(f):
        def callHelper(*args):
            for _ in range(0, times):
                f(*args)
        return callHelper
    return repeatHelper 
# fmt: on


def execute_with_probability(prob: float) -> bool:
    """
    Returns True of false based on the
    random variable prob

    Parameters
    ----------
    prob : float
        Probability to execute, from 0 (never happen)
        to 1 (always happen)

    Returns
    -------
    bool
        True prob % of the times
    """
    return np.random.uniform() > prob


# tensor/array operations


def to_numpy(x: Tensor) -> np.ndarray:
    """
    Converts a Tensor into a numpy array.

    Parameters
    ----------
    x : Tensor
        Original tensor

    Returns
    -------
    np.ndarray
        Converted np array
    """
    return x.cpu().detach().numpy()


def split_complex(x: TensorOrArray) -> TensorOrArray:
    """
    Splits a complex Tensor in a float
    Tensor with the double of the channels.

    Parameters
    ----------
    x : TensorOrArray
        Complex input of shape (B, C, ...)

    Returns
    -------
    TensorOrArray
        Float output of shape (B, 2*C, ...)
    """
    module = get_np_or_torch(x)
    x = module.cat((x.real, x.imag), 1)
    return x


def pack_complex(x: TensorOrArray) -> TensorOrArray:
    """
    Merges a 2*C channels float TensorOrArray in a complex TensorOrArray.

    Parameters
    ----------
    x : TensorOrArray
        Float input of shape (B, 2*C, ...)

    Returns
    -------
    TensorOrArray
        Complex output of shape (B, C, ...)
    """
    c = x.shape[1]
    x = x[:, : c // 2] + 1j * x[:, c // 2 :]
    return x



from typing import Iterator, List, Tuple, Union
from resampy import resample as resample_np
from torchaudio.functional import resample
from pathimport import set_module_root
from multiprocess import Pool
from itertools import islice
from pathlib import Path
from torch import Tensor
import soundfile as sf
import numpy as np
import torchaudio
import torch

set_module_root(".")
from torch_utilities.common import TensorOrArray
from torch_utilities.pytorch import get_device

# export list
__all__ = [
    "load_audio",
    "save_audio",
    "load_audio_parallel",
    "load_audio_parallel_itr",
]


def load_audio(
    file_path: Path,
    sample_rate: int = None,
    tensor: bool = False,
    device: str = "auto",
) -> Tuple[TensorOrArray, int]:
    """
    Loads an audio file.

    Parameters
    ----------
    file_path : Path
        Path to the audio file
    sample_rate : int, optional
        Target sample rate, by default avoids resample
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : str, optional
        The device to load the tensor into, by default auto

    Returns
    -------
    Tuple[np.ndarray, int]
        (audio of shape (C, T), sample_rate)
    """
    # audio loading
    if not tensor:
        data, old_sample_rate = sf.read(
            file_path,
        )
        data = data.T
        if len(data.shape) == 1:
            data = data[None, :]
    else:
        data, old_sample_rate = torchaudio.load(file_path)
        data = data.to(get_device() if device == "auto" else device)

    # resampling
    if sample_rate is None:
        sample_rate = old_sample_rate
    elif old_sample_rate != sample_rate:
        if tensor:
            data = resample(data, old_sample_rate, sample_rate)
        else:
            data = resample_np(data, old_sample_rate, sample_rate)

    return data, sample_rate


def save_audio(file_path: Path, data: TensorOrArray, sample_rate: int):
    """
    Saves an audio file

    Parameters
    ----------
    file_path : Path
        Path to the audio file
    data : TensorOrArray
        Audio file to save
    sample_rate : int, optional
        Target sample rate
    """
    dtype = type(data)

    if dtype == np.ndarray:
        sf.write(file_path, data.T, samplerate=sample_rate)
    elif dtype == Tensor:
        data = data.to("cpu")
        torchaudio.save(file_path, data, sample_rate=sample_rate)
    else:
        err_msg = f"{dtype} is not supported by save_audio"
        raise NotImplementedError(err_msg)


def load_audio_parallel(
    file_paths: List[Path],
    sample_rate: int = None,
    tensor: bool = False,
    device: str = "auto",
    num_workers: int = 4,
) -> List[TensorOrArray]:
    """
    Loads multiple audio files in parallel.

    Parameters
    ----------
    file_path : List[Path]
        List of paths to the audio files
    sample_rate : int, optional
        Target sample rate, by default None
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : str, optional
        The device to load the tensor into, by default auto
    num_workers : int, optional
        Number of parallel processes

    Returns
    -------
    List[TensorOrArray]
        list of audios of shape (C, T)
    """
    with Pool(num_workers) as pool:
        _load = lambda x: load_audio(x, sample_rate, False)[0]
        xs = pool.map(_load, file_paths)
    if tensor:
        xs = [Tensor(x).to(device=get_device() if device == "auto" else device) for x in xs]
    return xs


def load_audio_parallel_itr(
    file_paths: List[Path],
    sample_rate: int = None,
    tensor: bool = False,
    device: str = "auto",
    num_workers: int = 4,
) -> Iterator[TensorOrArray]:
    """
    Iterator that loads multiple audio files in parallel.

    Parameters
    ----------
    file_path : List[Path]
        List of paths to the audio files
    sample_rate : int, optional
        Target sample rate, by default None
    tensor : bool, optional
        If True loads a torch Tensor, by default False
    device : str, optional
        The device to load the tensor into, by default auto
    num_workers : int, optional
        Number of parallel processes

    Returns
    -------
    Iterator[TensorOrArray]
        yields audios of shape (C, T)
    """
    n_files = len(file_paths)
    # transforming to generator
    file_paths = (x for x in file_paths)
    for i in range(0, n_files, num_workers):
        files_batch = islice(file_paths, num_workers)
        cache = load_audio_parallel(files_batch, sample_rate, tensor, device, num_workers)
        for x in cache:
            yield x



from torch.utils.data import Sampler, Dataset, DataLoader, BatchSampler
from typing import List, Optional, Tuple
from pathimport import set_module_root
from pathlib import Path
from torch import Tensor
from tqdm import tqdm
import numpy as np
import itertools
import torch
import h5py

set_module_root(".")
from torch_utilities.common import to_numpy, split_complex

# export list
__all__ = [
    "WeakShufflingSampler",
    "HDF5Dataset",
    "get_hdf5_dataloader",
    "HDF5OnlineDataset",
    "get_dataset_statistics",
    "plot_dataset_statistics",
]


class WeakShufflingSampler(Sampler):
    def __init__(self, dataset: Dataset, batch_size: int):
        """
        Sampler that implements weak-shuffling.

        E.g. if dataset is [1,2,3,4] with batch_size=2,
             then first batch, [[1,2], [3,4]] then
             shuffle batches -> [[3,4],[1,2]]

        Referenece:
        https://towardsdatascience.com/reading-h5-files-faster-with-pytorch-datasets-3ff86938cc

        Parameters
        ----------
        dataset : Dataset
            Source dataset
        batch_size : int
            Size of the batches
        """
        self.batch_size = batch_size
        self.dataset_length = len(dataset)
        self.n_batches = int(np.ceil((self.dataset_length / self.batch_size)))
        self.batch_ids = torch.randperm(self.n_batches)

    def __len__(self):
        return self.batch_size

    def __iter__(self):
        datalen = self.dataset_length
        for id in self.batch_ids:
            start = id * self.batch_size
            end = (id + 1) * self.batch_size
            end = end if end < datalen else datalen
            selection = torch.arange(start, end)
            for index in selection:
                yield int(index)


class HDF5Dataset(Dataset):
    def __init__(
        self,
        dataset_path: Path,
        data_layout: List[str],
    ) -> None:
        """
        Dataset supporting HDF5 format.

        The following constraints should be satisfied
        by the HDF5 file for a correct behaviour:
        - Multiple single-level groups (e.g. /group1, /group2, ...)
        - The batch size (first dimension) of each dataset is the same
        - Each group has len(data_layout) dataset inside

        Parameters
        ----------
        dataset_path : Path
            Path to the .hdf5 file
        data_layout : List[str]
            List describing the layout of the data
            inside each group. For an input-label1-label2
            dataset the list would be ["input", "label1", "label2"]
        """
        super().__init__()
        self.dataset_path = dataset_path.absolute()
        self.data_layout = data_layout
        self.dataset_file = h5py.File(self.dataset_path, "r")
        self.groups = list(self.dataset_file.keys())
        self.group_batch_len = self.dataset_file[self.groups[0]][self.data_layout[0]].shape[0]
        self.data_layout = data_layout
        self._cache = None
        self._cache_idx = None

    def __len__(self):
        return len(self.groups) * self.group_batch_len

    def __getitem__(self, idx) -> List[Tensor]:
        # error handling
        err_msg = None
        if not isinstance(idx, list):
            err_msg = "HDF5Dataset must be sampled by a BatchLoader"
        elif len(idx) > self.group_batch_len:
            err_msg = "Reduce the batch size to be less than the batch size of the groups"
        elif self.group_batch_len % len(idx) != 0:
            err_msg = "Modify the batch size to be a divider of the HDF5 group batch size"
        if err_msg is not None:
            raise RuntimeError(err_msg)

        # cache update
        if not self._in_cache(idx[0]):
            self._update_cache(idx[0])

        # using the cache
        gbl = self.group_batch_len
        a, b = idx[0] % gbl, (idx[-1] % gbl) + 1
        data = [x[a:b] for x in self._cache]

        return data

    def _update_cache(self, idx: int) -> None:
        """
        Caches a group in memory.

        Parameters
        ----------
        idx : int
            Index of an element of the group
        """
        # getting the starting index of a group
        self._cache_idx = idx

        # updating the cache
        del self._cache
        g_idx = idx // self.group_batch_len
        g = self.dataset_file[self.groups[g_idx]]
        cast = lambda x: torch.from_numpy(np.array(x)).to("cpu")
        data = [cast(g[k]) for k in self.data_layout]
        self._cache = data

    def _in_cache(self, idx: int) -> bool:
        """
        Checks if an index is inside the cache.

        Parameters
        ----------
        idx : int
            Target index

        Returns
        -------
        bool
            True if the element is inside the cache
        """
        c_idx = self._cache_idx
        flag = (c_idx is not None) and (idx >= c_idx and idx < (c_idx + self.group_batch_len))
        return flag


def get_hdf5_dataloader(
    dataset_path: Path,
    data_layout: List[str],
    batch_size: int = None,
    **dataloader_kwargs: dict,
) -> DataLoader:
    """
    Create a dataloader binded to a HDF5Dataset.

    Parameters
    ----------
    dataset : Path
        Path to the HDF5 Dataset
    data_layout : List[str]
            List describing the layout of the data
            inside each group. For an input-label1-label2
            dataset the list would be ["input", "label1", "label2"]
    batch_size : int, optional
        Batch size of the dataloader, by default HDF5Dataset.group_batch_len
    dataloader_kwargs : dict, optional
        DataLoader arguments, by default {}
    """
    if batch_size is None:
        batch_size = dataset.group_batch_len

    if dataloader_kwargs is None:
        dataloader_kwargs = {}

    dataset = HDF5Dataset(dataset_path, data_layout)

    sampler = BatchSampler(
        sampler=WeakShufflingSampler(
            dataset=dataset,
            batch_size=batch_size,
        ),
        batch_size=batch_size,
        drop_last=False,
    )

    dataloader = DataLoader(
        dataset=dataset,
        sampler=sampler,
        **dataloader_kwargs,
    )

    return dataloader


class HDF5OnlineDataset(Dataset):
    def __init__(
        self,
        dataset_paths: List[Path],
        data_layouts: List[List[str]],
        batch_size: int = 16,
        total_items: int = 1024,
        overfit_mode: bool = False,
    ) -> None:
        """
        Dataset based on online generation.
        Data can be combined after retrieving it
        from source HDF5Datasets.

        Parameters
        ----------
        dataset_paths : List[Path]
            List of paths to the HDF5Datasets h5
        data_layouts : List[List[str]]
            Data layouts for all the HDF5Datasets
        batch_size : int, optional
            Batch size of each raw data, by default 16
        total_items : int, optional
            Total item in an epoch, by default 1024
        overfit_mode : bool, optional
            If True, always returns the first batch at every iteration
        """
        super(HDF5OnlineDataset).__init__()
        self.datasets_paths = dataset_paths
        self.data_layouts = data_layouts
        self.data_layouts = data_layouts
        self.batch_size = batch_size
        self.total_items = total_items
        self.overfit_mode = overfit_mode

        # error handling
        err_msg = f"total_items ({total_items}) must be divisible by batch_size ({batch_size})"
        assert (self.total_items % self.batch_size) == 0, err_msg

        # static datasets
        self.source_datasets = self._get_datasets()

        # handling overfit mode
        self.overfit_cache = None

    def __len__(self) -> int:
        return self.total_items // self.batch_size

    def transform(self, raw_data: List[Tensor]) -> List[Tensor]:
        """
        Override this method to process the raw data.

        Parameters
        ----------
        raw_data : List[Tensor]
            Raw data from the HDF5Datasets

        Returns
        -------
        List[Tensor]
            Processed data
        """
        return raw_data

    def _get_rand_batch(self, dataset: HDF5Dataset) -> List[Tensor]:
        """
        Get a random batch from an HDF5Dataset

        Parameters
        ----------
        dataset : HDF5Dataset
            Target HDF5Dataset

        Returns
        -------
        List[Tensor]
            Batches from the HDF5Datasets
        """
        # getting start and end points
        ds_len = len(dataset)
        total_batches = ds_len // self.batch_size
        r_int = np.random.randint(0, total_batches)
        start = r_int * self.batch_size
        end = (r_int + 1) * self.batch_size
        indices = list(range(start, end))

        # getting the random selection
        ds_outs = dataset[indices]
        return ds_outs

    def __getitem__(self, _) -> List[Tensor]:
        if self.overfit_cache is not None:
            # overfit cache enabled
            return self.overfit_cache

        outs = [self._get_rand_batch(ds) for ds in self.source_datasets]
        outs = itertools.chain.from_iterable(outs)
        outs = list(outs)
        outs = self.transform(outs)

        if self.overfit_mode:
            # caching for overfit mode
            self._store_in_cache(outs)

        return outs

    def _get_datasets(self) -> List[HDF5Dataset]:
        """
        Gets the static datasets.

        Returns
        -------
        List[HDF5Dataset]
            List of static datasets
        """
        ds = [HDF5Dataset(p, l) for p, l in zip(self.datasets_paths, self.data_layouts)]
        return ds

    def _store_in_cache(self, data: List[Tensor]) -> None:
        """
        Stores the data in the overfit cache.

        Parameters
        ----------
        data : List[Tensor]
            Data to store in the cache
        """
        self.overfit_cache = data


def get_dataset_statistics(
    dataloader: DataLoader,
    iterations: int,
    n_hist_bins: int = 100,
    verbose: bool = False,
) -> List[Tuple[np.ndarray, np.ndarray, float, float, float]]:
    """
    Calculate the statistics for the data obtained
    from a dataset.

    Parameters
    ----------
    dataloader : DataLoader
        DataLoader to iterate
    iterations : int
        Number of iterations
    n_hist_bins : int, optional
        Number of histogram bins
    verbose : bool, optional
        If True shows a progress bar for the iterations,
        by default False

    Returns
    -------
    List[Tuple[np.ndarray, np.ndarray, float, float, float, float]]
        List of statistics for each dataset output
        each tuple contains
        (hist_vals, hist_bins, min, max, mean, var)
    """
    ds = iter(dataloader)
    _range = tqdm(range(iterations)) if verbose else range(iterations)
    outs = [ds.__next__() for i in _range]
    n_items = len(outs[0])

    stats = []
    for i in range(n_items):
        x = torch.stack([o[i] for o in outs])
        x = x.flatten()
        x = x.float()
        x = split_complex(x) if torch.is_complex(x) else x
        hist_vals, hist_bins = [to_numpy(y) for y in x.histogram(n_hist_bins, density=True)]
        hist_vals /= n_hist_bins
        _min = x.min().item()
        _max = x.max().item()
        _mean = x.mean().item()
        _var = x.var().item()
        stats.append((hist_vals, hist_bins, _min, _max, _mean, _var))

    return stats


def plot_dataset_statistics(
    dataloader: DataLoader,
    iterations: int,
    n_hist_bins: int = 100,
    labels: Optional[List[str]] = None,
    figsize: Tuple[int, int] = (8, 6),
    verbose: bool = False,
) -> List[Tuple[np.ndarray, np.ndarray, float, float, float, float]]:
    """
    Plots the statistics for the data obtained
    from a dataset.

    Parameters
    ----------
    dataloader : DataLoader
        DataLoader to iterate
    iterations : int
        Number of iterations
    n_hist_bins : int, optional
        Number of histogram bins
    labels : Optional[List[str]]
        Names of the dataset outputs
    figsize : Tuple[int, int], optional
        Size of the plot, by default (8, 6)
    verbose : bool, optional
        If True shows a progress bar for the iterations,
        by default False

    Returns
    -------
    List[Tuple[np.ndarray, np.ndarray, float, float, float]]
        List of statistics for each dataset output
        each tuple contains
        (hist_vals, hist_bins, min, max, mean, var)
    """
    import matplotlib.pyplot as plt

    all_stats = get_dataset_statistics(
        dataloader,
        iterations,
        n_hist_bins,
        verbose,
    )

    if labels == None:
        # default labels
        labels = [f"feature_{i}" for i in range(len(all_stats))]

    for stats, name in zip(all_stats, labels):
        hist_vals, hist_bins, min, max, mean, var = stats
        bar_width = np.abs(hist_bins[1] - hist_bins[0])
        plt.figure(figsize=figsize)
        plt.title(f"{name} distribution")
        plt.bar(hist_bins[:-1], hist_vals, width=bar_width)
        plt.vlines(x=mean, ymin=0, ymax=1, colors="red", label="mean")
        plt.ylim([0, hist_vals.max()])
        plt.legend()
        plt.grid()
        plt.show()
        plt.close()

    return all_stats



from pathimport import set_module_root
import numpy as np
import unittest
import torch

set_module_root("../torch_utils")
import torch_utilities as tu
import torch_utilities.augmentation as aug
from torch_utilities.audio import invert_db


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestAugmentation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass

    def test_shuffle(self):
        torch.manual_seed(0)
        x = torch.arange(1000)
        y = aug.shuffle(x)
        self.assertEqual(x.shape, y.shape)
        self.assertTrue(torch.any(x.not_equal(y)))

    def test_add_noise(self):
        snr_set = (-12, 0, 12)
        for snr in snr_set:
            with self.subTest(snr=snr):
                x = torch.ones((1, 1, 100))
                n = torch.ones_like(x)
                y, actual_snr = aug.add_noise(x, n, (snr, snr))
                self.assertTrue(x.max().equal(y.max()))
                self.assertLess(actual_snr - invert_db(snr), 1e-6)
                self.assertTrue(len(actual_snr.shape), 1)
                self.assertTrue(actual_snr.shape[0], y.shape[0])

    def test_scale(self):
        scale_set = (-12, 0, 12)
        for scale in scale_set:
            with self.subTest(snr=scale):
                lin_scale = aug.invert_db(scale)
                x = torch.ones((1, 1, 100))
                y, actual_scaling = aug.random_scaling(x, (scale, scale))
                self.assertAlmostEqual(y.max().item(), lin_scale)
                self.assertLess(actual_scaling - invert_db(scale), 1e-6)
                self.assertTrue(len(actual_scaling.shape), 1)
                self.assertTrue(actual_scaling.shape[0], y.shape[0])

    def test_overdrive(self):
        x = torch.ones((1, 1, 100))
        y = aug.random_overdrive(x)
        self.assertAlmostEqual(y.max().item(), 1)

    def test_dc_removal(self):
        x = torch.ones((2, 3, 5))
        y = aug.dc_removal(x)
        m = y.mean()
        self.assertLess(m, 1e-6)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from pathimport import set_module_root

set_module_root(".")
from torch_utilities.io import *
from torch_utilities.audio import *
from torch_utilities.common import *
from torch_utilities.pytorch import *
from torch_utilities.modules import *
from torch_utilities.data_loading import *
from torch_utilities.augmentation import *
from torch_utilities.model_trainer import *



from typing import Tuple
from pathimport import set_module_root
from ray import tune
from torch import nn
import numpy as np
import itertools
import unittest
import torch

set_module_root("../torch_utils")
set_module_root(".")
import torch_utilities as tu
from tests.generate_test_data import get_test_data_dir


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.config_path = get_test_data_dir() / "test.yml"
        self.config = tu.Config(self.config_path)

    def test_get_str(self):
        y_true = "test"
        y_test = self.config.get("section1", "param1", str)
        self.assertEqual(y_true, y_test)

    def test_get_int(self):
        y_true = 42
        y_test = self.config.get("section1", "param2", int)
        self.assertEqual(y_true, y_test)

    def test_get_list(self):
        y_true = ["a", "b", "c"]
        y_test = self.config.get("section1", "param3", list)
        self.assertEqual(y_true, y_test)

    def test_get_float(self):
        y_true = 12.43
        y_test = self.config.get("section2", "param4", float)
        self.assertAlmostEqual(y_true, y_test)

    def test_default_none(self):
        y_true = None
        y_test = self.config.get("section1", "param10")
        self.assertEqual(y_true, y_test)

    def test_default(self):
        y_true = "default"
        y_test = self.config.get("section1", "param10", default="default", _type=int)
        self.assertEqual(y_true, y_test)

    def test_get_ray_tune_params(self):
        params = self.config.get_ray_tune_params()
        sampling_methods = set(params.keys())
        params_names = ["weight_decay", "loss_weight_0", "depth", "learning_rate", "list_choice"]
        self.assertEqual(sampling_methods, set(params_names))
        self.assertEqual(type(params), dict)
        params = tu.DotDict(params)

        # random variable type
        self.assertEqual(type(params.weight_decay), tune.search.sample.Float)
        self.assertEqual(type(params.learning_rate), tune.search.sample.Float)
        self.assertEqual(type(params.loss_weight_0), tune.search.sample.Float)
        self.assertEqual(type(params.depth), tune.search.sample.Categorical)
        self.assertEqual(type(params.list_choice), tune.search.sample.Categorical)

        # type after sampling
        self.assertEqual(type(params.weight_decay.sample()), float)
        self.assertEqual(type(params.learning_rate.sample()), float)
        self.assertEqual(type(params.loss_weight_0.sample()), float)
        self.assertEqual(type(params.depth.sample()), int)
        self.assertEqual(type(params.list_choice.sample()), list)


class TestGeneric(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass

    def test_pack_many(self):
        xs = [1, 2, 3]
        ys = [4, 5, 6]
        zss = tu.pack_many(xs, ys)
        self.assertEqual(zss, [(1, 4), (2, 5), (3, 6)])

    def test_one_hot_quantization(self):
        min_max = ((-1, 1), (0, 65535), (11.1, 23.9))
        steps = (8, 256, 1024)
        params = itertools.product(min_max, steps)
        for p in params:
            (min_, max_), s = p
            with self.subTest(p=p):
                C, T, F = 2, 10, 32
                x = torch.rand(1, C, T, F) * (max_ - min_) + min_
                y = tu.one_hot_quantization(x, s, min_, max_)
                self.assertEqual(y.shape, (1, s, C, T, F))
                self.assertEqual(y.sum().item(), C * T * F)

    def test_invert_one_hot(self):
        n_dims = (2, 3, 4)
        steps = (3, 10, 256)
        params = itertools.product(n_dims, steps)
        for d, s in params:
            with self.subTest(d=d, s=s):
                x = torch.rand([3] * d) * s
                x = torch.floor(x).clip(0, s - 1)
                y = tu.one_hot_quantization(x, s, 0, s)
                x_hat = tu.invert_one_hot(y)
                self.assertTrue(x.equal(x_hat))


class TestTensorArrayOperations(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass

    def test_split_complex(self):
        mod = (np, torch)
        chl = (1, 2, 4, 8)
        params = itertools.product(mod, chl)
        xs = [m.ones((1, c, 16), dtype=m.complex64) for m, c in params]
        for x, p in zip(xs, params):
            with self.subTest(p=p):
                y = tu.split_complex(x)
                c = y.shape[1]
                _ones = mod.ones_like(y[:, : c // 2])
                _zeros = mod.zeros_like(y[:, : c // 2])
                self.assertEqual(c, x.shape[1] * 2)
                self.assertTrue(mod.allclose(y[:, : c // 2]), _ones)
                self.assertTrue(mod.allclose(y[:, c // 2 :]), _zeros)

    def test_pack_complex(self):
        mod = (np, torch)
        chl = (2, 4, 8)
        params = itertools.product(mod, chl)
        xs = [m.ones((1, c, 16)) for m, c in params]
        for x in xs:
            with self.subTest(x=x):
                y = tu.pack_complex(x)
                self.assertEqual(y.shape[1], x.shape[1] // 2)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from pathimport import set_module_root
import onnxruntime as ort
from typing import Tuple
from pathlib import Path
from torch import Tensor
import numpy as np


set_module_root(".")
from torch_utilities.common import to_numpy, TensorOrArray
from torch_utilities.audio import MelFilterbank, stft

# export list
__all__ = ["DNSMOS"]


def _get_model_dir() -> Path:
    """
    Returns the directory containing the models.

    Returns
    -------
    Path
        Path to the directory
    """
    path = Path(__file__).parent.parent / "resources" / "models"
    return path


class DNSMOS:
    def __init__(self) -> None:
        self.model_path = _get_model_dir() / "sig_bak_ovr.onnx"
        self.inference_sess = self.get_inference_session()
        self.sample_rate = 16000
        self.hopsize_ms = 10
        self.n_mels = 120
        self.n_freqs = int(self.sample_rate / 1000 * self.hopsize_ms * 2 + 1)

        # transforms
        self.stft = lambda x: stft(
            x,
            sample_rate=self.sample_rate,
            hopsize_ms=self.hopsize_ms,
            win_oversamp=2,
        )
        self.mel_fb = MelFilterbank(
            sample_rate=self.sample_rate,
            n_freqs=self.n_freqs,
            n_mels=self.n_mels,
        )

    def get_inference_session(self) -> ort.InferenceSession:
        """
        Returns an inference session for the model.

        Returns
        -------
        ort.InferenceSession
            DNSMOS inference session
        """
        session = ort.InferenceSession(str(self.model_path))
        return session

    def get_polyfit_val(self, ovr: float, sig: float, bak: float) -> Tuple[float, float, float]:
        """
        Polynomial nonlinearity as in the original repo

        Parameters
        ----------
        ovr : float
            Ovr raw value
        sig : float
            Sig raw value
        bak : float
            Bak raw value

        Returns
        -------
        Tuple[float, float, float]
            Processed metrics
        """
        p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
        p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
        p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        ovr_poly = p_ovr(ovr)
        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)

        return ovr_poly, sig_poly, bak_poly

    def __call__(self, x: TensorOrArray) -> Tuple[float, float, float]:
        """
        Compute ovr, sig and bak from an 16kHz input.

        Parameters
        ----------
        x : TensorOrArray
            Input of shape (C, T) sampled at 16kHz,
            if multichannel, the first channel only is considered

        Returns
        -------
        Tuple[float, float, float]
            (ovr, sig, bak)
        """
        assert len(x.shape) == 2, "the shape of x should be (C, T)"

        # # Mel transform
        # x = self.stft(x)
        # x = self.mel_fb(x)
        if isinstance(x, Tensor):
            x = to_numpy(x)
        x = x.astype("float32")

        # keeping the first channel
        x = x[:1]

        # computing the average over segments of 1s
        samples_selected = 144160
        n_frames = x.shape[1] // samples_selected
        ovr = []
        sig = []
        bak = []
        for i in range(n_frames):
            a = i * samples_selected
            b = (i + 1) * samples_selected
            x_sel = x[:, a:b]
            model_in = dict(input_1=x_sel)
            scores = self.inference_sess.run(None, model_in)[0][0]
            scores = self.get_polyfit_val(*scores)
            [container.append(s) for container, s in zip((sig, bak, ovr), scores)]
        ovr, sig, bak = [np.mean(container) for container in (ovr, sig, bak)]

        return ovr, sig, bak



from pathimport import set_module_root
from itertools import product
from torch import Tensor
import numpy as np
import unittest
import torch

set_module_root("../torch_utils")
from tests.generate_test_data import get_test_data_dir
import torch_utilities as tu


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestIO(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.mono_file = get_test_data_dir() / "mono.wav"
        self.stereo_file = get_test_data_dir() / "stereo.wav"
        self.tmp_file = get_test_data_dir() / "tmp.wav"

    def delete_tmp(self):
        self.tmp_file.unlink(missing_ok=True)

    def test_load_audio(self):
        mono = (False, True)
        sample_rate = (None, 16000, 48000)
        tensor = (False, True)
        device = ("cpu", "auto")
        params = product(
            mono,
            sample_rate,
            tensor,
            device,
        )
        for p in params:
            (m, sr, t, d) = p
            default_sr = 16000
            with self.subTest(p=p):
                file = self.mono_file if m else self.stereo_file
                x, x_sr = tu.load_audio(file, sr, t, d)
                self.assertEqual(len(x.shape), 2)
                self.assertEqual(x.shape[0], 1 if m else 2)
                self.assertEqual(type(x), Tensor if t else np.ndarray)
                self.assertEqual(x_sr, sr if sr is not None else default_sr)
                if t:
                    expected_device = "cpu" if d == "cpu" else tu.get_device()
                    self.assertEqual(x.device.type, expected_device)

    def test_save_audio(self):
        mono = (False, True)
        sample_rate = (16000, 48000)
        tensor = (False, True)
        params = product(
            mono,
            sample_rate,
            tensor,
        )
        for p in params:
            (m, sr, t) = p
            with self.subTest(p=p):
                ch = 1 if m else 2
                x = torch.zeros((ch, sr)) if t else np.zeros((ch, sr))
                tu.save_audio(self.tmp_file, x, sr)
                self.delete_tmp()

    def test_load_audio_parallel(self):
        mono = (False, True)
        sample_rate = (16000, 48000)
        tensor = (False, True)
        device = ("cpu", "auto")
        num_workers = (1, 4)
        params = product(
            mono,
            sample_rate,
            tensor,
            device,
            num_workers,
        )
        for p in params:
            (m, sr, t, d, w) = p
            n_files = 4
            with self.subTest(p=p):
                name = "mono.wav" if m else "stereo.wav"
                filepaths = [get_test_data_dir() / name for _ in range(n_files)]
                xs = tu.load_audio_parallel(filepaths, sr, t, d, w)
                self.assertEqual(len(xs), n_files)
                for x in xs:
                    self.assertEqual(type(x), Tensor if t else np.ndarray)
                    self.assertEqual(x.shape[0], 1 if m else 2)

    def test_load_audio_parallel_itr(self):
        mono = (False, True)
        sample_rate = (16000, 48000)
        tensor = (False, True)
        device = ("cpu", "auto")
        num_workers = (1, 2)
        params = product(
            mono,
            sample_rate,
            tensor,
            device,
            num_workers,
        )
        for p in params:
            (m, sr, t, d, w) = p
            n_files = 4
            with self.subTest(p=p):
                name = "mono.wav" if m else "stereo.wav"
                filepaths = [get_test_data_dir() / name for _ in range(n_files)]
                xs = tu.load_audio_parallel_itr(filepaths, sr, t, d, w)
                for i, x in enumerate(xs):
                    self.assertEqual(type(x), Tensor if t else np.ndarray)
                    self.assertEqual(x.shape[0], 1 if m else 2)
                self.assertEqual(i + 1, n_files)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from torch.utils.tensorboard import SummaryWriter
from contextlib import suppress, nullcontext
from torch.utils.data import DataLoader
from pathimport import set_module_root
import matplotlib.pyplot as plt
from torch.optim import Optimizer
from torch import nn, Tensor
import torch_utilities as tu
from loguru import logger
from pathlib import Path
from abc import ABC
import numpy as np
import warnings
import torch
import math
import yaml
import abc

set_module_root(".")
from torch_utilities.common import DotDict
from torch_utilities.pytorch import load_checkpoints, sort_checkpoints

__all__ = ["ModelTrainer"]


class ModelTrainer(ABC):
    def __init__(
        self,
        model_path: Path,
        model: nn.Module,
        train_dl: DataLoader,
        valid_dl: DataLoader,
        optimizer_class: Type[Optimizer],
        losses: List[Callable],
        net_ins_indices: Optional[List[int]] = None,
        losses_names: Optional[List[str]] = None,
        save_buffer_maxlen: int = 10,
        gradient_clip_value: Optional[float] = None,
        enable_profiling: bool = False,
    ) -> None:
        """
        Abstract class to structure a model training.

        The directory pointed by model_path should
        have the following structure:

        .
        ├── config.yml
        ├── checkpoints  // optional, from previous trainings
        └── logs         // optional, from previous trainings

        Use the section [training] of config.yml to store
        additional parameters

        Parameters
        ----------
        model_path : Path
            Path to the model directory
        model : nn.Module
            Model to train
        train_dl : DataLoader
            Training DataLoader
        train_dl : DataLoader
            Validation DataLoader
        optimizer_class : Type[Optimizer]
            Optimizer class
        losses : List[Callable]
            List of losses to be computed
        net_ins_indices : Optional[List[int]]
            Used to filter and reorder the List[Tensor] obtained
            from the dataloader, by default no filter.
            E.g.
                dataloader out = [T0, T1, T2]
                net_ins_indices = [2, 1]
                net_ins = [T2, T1]
        losses_names : Optional[List[str]]
            Names of the losses, by default ["loss0", "loss1", ...]
        save_buffer_maxlen : int, optional
            N best checkpoints saved (based on a lower total loss),
            by default 5
        gradient_clip_value : Optional[float]
            Maximum gradient update value, by default no gradient clipping
        enable_profiling : bool, optional
            If True runs one full run of the train dataset and
            logs to tensorboard the profiling information, by default False


        config.yml [training] parameters
        ----------
        max_epochs : int, optional
            Max number of epochs, by default 100
        learning_rate : float, optional
            Optimizer learning rate starting value, by default 1e-5
        overfit_mode : bool, optional
            Enables overfit mode, by default False
        weight_decay : float, optional
            Optimizer weight_decay, by default 0
        losses_weights : List[float], optional
            Per-loss gains, by default all ones
        log_every : int, optional
            Frequency of the logs (over the dataset iterations),
            by default 100
        """
        self.device = tu.get_device()

        # explicit attributes
        self.model_path = model_path
        self.model = model
        self.train_dl = train_dl
        self.valid_dl = valid_dl
        self.losses = losses
        self.net_ins_indices = net_ins_indices or [0]
        self.losses_names = losses_names or self._default_losses_names()
        self.optimizer_class = optimizer_class
        self.gradient_clip_value = gradient_clip_value
        self.enable_profiling = enable_profiling

        # configuration attributes
        self.config_path = self.model_path / "config.yml"
        self.config = tu.Config(self.config_path)
        self.learning_rate = self._from_config("learning_rate", float, 1e-5)
        self.weight_decay = self._from_config("weight_decay", float, 0)
        self.log_every = self._from_config("log_every", int, 100)
        self.max_epochs = self._from_config("max_epochs", int, 100)
        self.overfit_mode = self._from_config("overfit_mode", bool, False)
        self.losses_weights = self._from_config(
            "losses_weights", np.array, np.ones(len(self.losses))
        )
        self.losses_weights = self.losses_weights.astype(float)

        # other dirs
        self.checkpoints_dir = model_path / "checkpoints"
        self.checkpoint_monitoring_file = self.checkpoints_dir / "ckpt.yml"
        self.logs_dir = model_path / "logs"
        [d.mkdir(exist_ok=True) for d in (self.checkpoints_dir, self.logs_dir)]

        # model and running_losses setup
        self.start_epoch = 0
        self.save_buffer = []
        self.save_buffer_maxlen = save_buffer_maxlen
        self.net = self._load_model()
        self.optim_state = None
        self.optimizer = self._setup_optimizer()
        self.disable_optimization = False
        self.running_losses = None
        self.running_losses_steps = 0
        self.total_loss_name = "total"
        self._reset_running_losses()

        # extra stuff
        self.is_validation = False
        self.log_writer = SummaryWriter(self.logs_dir)
        self.figsize = (8, 6)
        self.dummy_input_train = self._get_dummy_data(True)
        self.dummy_input_valid = self._get_dummy_data(False)
        self.last_total_loss = 1e10
        self.last_computed_metric = -1e10  # used to select the best checkpoints
        self.profiler = self._get_profiler()  # null context manager if enable_profiling==False

    # = = = = = = = = = = = = = = = = = = = = = =
    #             Getters/Setters
    # = = = = = = = = = = = = = = = = = = = = = =
    def get_model(self) -> nn.Module:
        """
        Gets the model instance.

        Returns
        -------
        nn.Module
            Model instance
        """
        return self.net

    def is_training(self) -> bool:
        """
        Returns True if the train steps are running.

        Returns
        -------
        bool
            True if the train steps are running.
        """
        return not self.is_validation

    def get_losses(self) -> List[Callable]:
        """
        Gets the loss functions.

        Returns
        -------
        List[Callable]
            List of loss functions in the same order as
            they're passed to the constructor
        """
        return self.losses

    def get_losses_weights(self) -> List[float]:
        """
        Gets the weight of losses.

        Returns
        -------
        List[float]
            List of loss weights in the same order as
            they're passed to the constructor
        """
        return self.losses_weights

    def get_learning_rate(self) -> float:
        """
        Gets the current learning rate.

        Returns
        -------
        float
            Current learning rate
        """
        for param_group in self.optimizer.param_groups:
            return param_group["lr"]

    def set_learning_rate(self, value: float) -> None:
        """
        Setter for learning_rate.

        Parameters
        ----------
        value : float
            New learning rate
        """
        for g in self.optimizer.param_groups:
            g["lr"] = value

    def set_weight_decay(self, value: float) -> None:
        """
        Setter for weight_decay.

        Parameters
        ----------
        value : float
            New learning rate
        """
        for g in self.optimizer.param_groups:
            g["weight_decay"] = value

    # = = = = = = = = = = = = = = = = = = = = = =
    #             Training loop
    # = = = = = = = = = = = = = = = = = = = = = =
    def start_training(self) -> None:
        """
        Trains a model.
        """
        # logging
        logger.info(f"device selected: {self.device}")
        logger.info(f"model: {self.model_path.name}")
        logger.info(f"parameters: {self._get_model_parameters() / 1e3} K")
        if self.overfit_mode:
            logger.info("overfit mode on")
        if self.enable_profiling:
            logger.info("profiler on: stopping after one epoch")
        else:
            logger.info("learning rate scheduler enabled")
        logger.info("saving the model graph")
        self._log_graph()
        logger.info("saving the text of config.yaml")
        self._log_yaml()

        msg = "starting profilation" if self.enable_profiling else "starting training"

        logger.info(msg)  # - = - § >>
        self._start_profiling()  # <- - profiler on
        self.on_train_begin()
        for epoch in range(self.start_epoch, self.max_epochs):
            self.on_epoch_begin(epoch)
            logger.info(f"epoch [{epoch+1}/{self.max_epochs}]")

            # training
            self.net.train()
            self.is_validation = False
            for i, data in enumerate(self.train_dl):
                data = self._remove_extra_dim(data)
                # in overfit mode dummy data is used instead of real data
                data = self.dummy_input_train if self.overfit_mode else data
                self._train_step(data, epoch)
                if i % self.log_every == 0 and i != 0:
                    logger.info(f"batch [{i}/{len(self.train_dl)}]")
                    self._log_losses(epoch=epoch)
                #
            self._log_gradients(epoch)
            self._log_losses(epoch=epoch)

            _log_data = lambda t: self.dummy_input_train if t else self.dummy_input_valid
            with torch.no_grad():
                logger.info("logging tensorboard train data")
                self.tensorboard_logs(_log_data(True), epoch=epoch)
                self._log_outs(epoch)

                # leaving the training  ~ ~
                self._stop_profiling()  # <- - profiler off
                if self.enable_profiling:
                    logger.info(f"{self.model_path.name} profilation complete")
                    return
                # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

                if not self.overfit_mode:
                    # validation
                    self.net.eval()
                    self.is_validation = True
                    for i, data in enumerate(self.valid_dl):
                        data = self._remove_extra_dim(data)
                        self._valid_step(data, epoch)
                    self._log_losses(epoch=epoch)
                    logger.info("logging tensorboard valid data")
                    self.tensorboard_logs(_log_data(False), epoch=epoch)

            metric_log_data = _log_data(self.is_training())
            self.last_computed_metric = self.apply_metric(metric_log_data)
            self._save_model(epoch)
            self.on_epoch_end(epoch)
            self._check_nan()

        logger.info(f"{self.model_path.name} training complete")  # - = - § >>
        self.on_train_end()

    def _train_step(self, data: List[Tensor], epoch: int) -> None:
        """
        Single step of a training loop.

        Parameters
        ----------
        data : List[Tensor]
            Inputs to the model
        epoch : int
            Current epoch
        """
        self.on_train_step_begin(epoch)
        data = [x.to(self.device) for x in data]
        net_ins = self._get_filtered_input(data)
        self.optimizer.zero_grad()
        net_outs = self.net(*net_ins)
        _losses = self.apply_losses(data, net_outs)
        _losses = self._apply_losses_weights(_losses)

        # enable custom optimization schemes ~ ~ ~ ~ ~ ~
        if not self.disable_optimization:
            total_loss = sum(_losses)
            total_loss.backward()
            # gradient clipping
            if self.gradient_clip_value is not None:
                nn.utils.clip_grad_value_(
                    self.net.parameters(),
                    clip_value=self.gradient_clip_value,
                )
            self.optimizer.step()
            # for logging
            self._update_running_losses(_losses)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

        self.on_train_step_end(epoch)

        # profiler update
        if self.enable_profiling:
            self.profiler.step()

    def _valid_step(self, data: List[Tensor], epoch) -> None:
        """
        Single step of a validation loop.

        Parameters
        ----------
        data : List[Tensor]
            Inputs to the model
        epoch : int
            Current epoch
        """
        self.on_valid_step_begin(epoch)
        data = [x.to(self.device) for x in data]
        net_ins = [data[i] for i in self.net_ins_indices]
        net_outs = self.net(*net_ins)
        _losses = self.apply_losses(data, net_outs)
        _losses = self._apply_losses_weights(_losses)
        self._update_running_losses(_losses)
        self.on_valid_step_end(epoch)

    # = = = = = = = = = = = = = = = = = = = = = =
    #            Handling Losses
    # = = = = = = = = = = = = = = = = = = = = = =
    @abc.abstractmethod
    def apply_losses(self, data: List[Tensor], net_outs: List[Tensor]) -> List[Tensor]:
        """
        Use the new_outputs to calculate the losses and
        return them.

        Parameters
        ----------
        data : List[Tensor]
            List of Tensors loaded by the dataloader
        net_outs : List[Tensor]
            Network outputs

        Returns
        -------
        List[Tensor]
            List of computed losses (not weighted)
        """
        pass

    def apply_metric(self, data: List[Tensor]) -> float:
        """
        Defines the metric used to keep the best checkpoints.
        The higher the better.

        Parameters
        ----------
        data : List[Tensor]
            List of Tensors loaded by the dataloader

        Returns
        -------
        float
            Value of the metric, the higher the better
        """
        # by default the metric is the inverse of the last_total_loss
        return -float(self.last_total_loss)

    def _apply_losses_weights(
        self,
        losses: List[Union[Tensor, float]],
    ) -> List[Union[Tensor, float]]:
        """
        Weights the losses by losses_weights.

        Parameters
        ----------
        losses : List[Union[Tensor, float]]
            List of losses to weight

        Returns
        -------
        Tensor
            Weighted losses
        """
        return [loss * w for loss, w in zip(losses, self.losses_weights)]

    def _update_running_losses(self, losses: List[Tensor]) -> None:
        """
        Updates the running losses.

        Parameters
        ----------
        losses : List[Tensor]
            Current losses
        """
        losses = [loss.item() for loss in losses]
        self.running_losses = [l0 + l1 for l0, l1 in zip(self.running_losses, losses)]
        self.running_losses_steps += 1

    def _reset_running_losses(self) -> None:
        """
        Resets the state of the running_losses.
        """
        self.running_losses = np.zeros(len(self.losses))
        self.running_losses_steps = 0

    # = = = = = = = = = = = = = = = = = = = = = =
    #               Callbacks
    # = = = = = = = = = = = = = = = = = = = = = =
    def on_train_begin(self: int) -> None:
        pass

    def on_train_end(self: int) -> None:
        pass

    def on_epoch_begin(self, epoch: int) -> None:
        pass

    def on_epoch_end(self, epoch: int) -> None:
        pass

    def on_train_step_begin(self, epoch: int) -> None:
        pass

    def on_train_step_end(self, epoch: int) -> None:
        pass

    def on_valid_step_begin(self, epoch: int) -> None:
        pass

    def on_valid_step_end(self, epoch: int) -> None:
        pass

    # = = = = = = = = = = = = = = = = = = = = = =
    #         Model loading/saving
    # = = = = = = = = = = = = = = = = = = = = = =
    def _load_model(self) -> nn.Module:
        """
        Loads a model from its class and configuration file.

        Returns
        -------
        nn.Module
            Loaded model
        """
        m = self.model(self.config_path)
        m.to(self.device)

        # load checkpoint if it exists
        if self._prev_train_exists():
            epoch, model_state, optim_state = self._load_checkpoint()
            self.start_epoch = epoch
            self.optim_state = optim_state
            m.load_state_dict(model_state)

        return m

    def _prev_train_exists(self) -> bool:
        """
        Checks if a previous training exists.

        Returns
        -------
        bool
            True if a previous training exists
        """
        if not self.checkpoints_dir.exists():
            return False
        checkpoints = self._get_checkpoints()
        return len(checkpoints) > 0

    def _get_checkpoints(self) -> List[Path]:
        """
        Gets the checkpoints paths.

        Returns
        -------
        List[Path]
            Checkpoints paths
        """
        checkpoints = list(self.checkpoints_dir.glob("*.ckpt"))
        return checkpoints

    def _load_checkpoint(self) -> Tuple[int, Dict, Dict, Dict]:
        """
        Loads the best checkpoint.

        Returns
        -------
        Tuple[int, Dict, Dict]
            (epoch, model_state, optim_state)
        """
        # choosing the best checkpoint
        self._load_checkpoint_monitoring()
        best_checkpoint = self.save_buffer[0][0]
        best_checkpoint = self.checkpoints_dir / best_checkpoint

        # loading and parsing
        best_checkpoint = DotDict(torch.load(best_checkpoint, map_location=self.device))
        epoch = best_checkpoint.epoch + 1
        model_state = best_checkpoint.model_state
        optim_state = best_checkpoint.optim_state
        return epoch, model_state, optim_state

    def _save_model(self, epoch: int) -> None:
        """
        Saves the model in the checkpoints folder.

        Parameters
        ----------
        epoch : int
            Current epoch
        """
        # saving
        checkpoint_path = self.checkpoints_dir / f"checkpoint_{epoch+1}.ckpt"
        torch.save(
            dict(
                model_state=self.net.state_dict(),
                optim_state=self.optimizer.state_dict(),
                epoch=epoch,
            ),
            checkpoint_path,
        )
        self._push_new_checkpoint(checkpoint_path.name)

    def _push_new_checkpoint(self, checkpoint_name: str) -> None:
        """
        Pushes a new checkpoint into the save_buffer and keeps track
        of the best ones.

        Parameters
        ----------
        checkpoint_name : str
            Name of the checkpoint
        """
        # sorting by the highest metric
        self.save_buffer.append((checkpoint_name, float(self.last_computed_metric)))
        self._sort_checkpoints()

        self._delete_worse_checkpoints()
        self._save_checkpoint_monitoring()

    def _save_checkpoint_monitoring(self) -> None:
        """
        Saves the state of the save_buffer.
        """
        with open(self.checkpoint_monitoring_file, "w") as f:
            yaml.dump(self.save_buffer, f)

    def _load_checkpoint_monitoring(self) -> None:
        """
        Loads the checkpoints.
        """
        self.save_buffer = load_checkpoints(self.checkpoint_monitoring_file)

    def _sort_checkpoints(self) -> None:
        """
        Sorts the save_buffer based on the metric value.
        """
        self.save_buffer = sort_checkpoints(self.save_buffer)
        self.save_buffer = self.save_buffer[: self.save_buffer_maxlen]

    def _delete_worse_checkpoints(self) -> None:
        """
        Deletes the checkpoints that are not in the save_buffer.
        """
        sb = [t[0] for t in self.save_buffer]
        checkpoints = self.checkpoints_dir.glob("*.ckpt")
        targets = filter(lambda x: x.name not in sb, checkpoints)
        [x.unlink() for x in targets]

    # = = = = = = = = = = = = = = = = = = = = = =
    #           Optimizer loading
    # = = = = = = = = = = = = = = = = = = = = = =
    def _setup_optimizer(self) -> Optimizer:
        """
        Sets up the optimizer to the model.

        Returns
        -------
        Optimizer
            Optimizer instance
        """
        try:
            optim = self.optimizer_class(
                self.net.parameters(),
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
        except TypeError:
            # no weight decay
            optim = self.optimizer_class(
                self.net.parameters(),
                lr=self.learning_rate,
            )

        if self.optim_state is not None:
            with suppress(RuntimeError):
                # ignore errors for optimizer mismatches
                optim.load_state_dict(self.optim_state)
        return optim

    # = = = = = = = = = = = = = = = = = = = = = =
    #                Logging
    # = = = = = = = = = = = = = = = = = = = = = =
    def get_tensorboard_writer(self) -> SummaryWriter:
        """
        Getter to the tensorboard log writer

        Returns
        -------
        SummaryWriter
            tensorboard log writer
        """
        return self.log_writer

    @abc.abstractclassmethod
    def tensorboard_logs(self, raw_data: List[Tensor], epoch: int) -> None:
        """
        Additional tensorboard logging.

        Parameters
        ----------
        raw_data : List[Tensor]
            Dataset input
        epoch : int
            Current epoch
        """
        pass

    def _get_dummy_data(self, is_training: bool) -> List[Tensor]:
        """
        Returns an input from the dataset

        Parameters
        -------
        is_training : bool
            Flag to separate train/valid logging

        Returns
        -------
        List[Tensor]
            Dataset input
        """
        ds = self.train_dl if is_training else self.valid_dl
        data = [x.to(self.device) for x in ds.dataset[[0, 1]]]
        return data

    def _log_losses(self, epoch: int) -> None:
        """
        Logs running_losses.

        Parameters
        ----------
        epoch : int
            Current epoch
        """
        if self.running_losses_steps == 0:
            # nothing to log
            return

        tag_suffix = "train" if self.is_training() else "valid"
        losses_names = self.losses_names + [self.total_loss_name]
        total_loss = sum(self.running_losses)
        losses = self.running_losses + [total_loss]
        for loss, name in zip(losses, losses_names):
            loss /= self.running_losses_steps
            logger.info(f"{name}: {loss}")
            self.log_writer.add_scalar(f"{name}_{tag_suffix}", loss, global_step=epoch)

        self.last_total_loss = total_loss / self.running_losses_steps
        self._reset_running_losses()

    def _default_losses_names(self) -> List[str]:
        """
        Default loss names

        Returns
        -------
        List[str]
            ["loss0", "loss1", ...]
        """
        return [f"loss{i}" for i in range(len(self.losses))]

    def _log_gradients(self, epoch: int) -> None:
        """
        Saves a plot of the gradient norm in tensorboard.

        Parameters
        ----------
        epoch : int
            Current epoch
        """
        grad = tu.get_gradients(self.net)
        plt.figure(figsize=self.figsize)
        plt.bar(range(grad.shape[0]), grad)
        plt.title("model gradients")
        plt.grid()
        plt.xlabel("submodule index")
        plt.ylabel("gradient norm")
        fig = plt.gcf()
        self.log_writer.add_figure("gradient", fig, epoch)
        plt.close()

    def _log_outs(self, epoch: int) -> None:
        """
        Saves a plot of the outputs norm in tensorboard.

        Parameters
        ----------
        epoch : int
            Current epoch
        """
        net_ins = self._get_filtered_input(self.dummy_input_train)
        net_outs = self.net(*net_ins)
        net_outs = [x.flatten().cpu() for x in net_outs]

        plt.figure(figsize=self.figsize)
        plt.violinplot(net_outs)
        plt.title("model outputs")
        plt.grid()
        plt.xlabel("output index")
        # plt.ylabel("")
        fig = plt.gcf()
        self.log_writer.add_figure("outputs", fig, epoch)
        plt.close()

    def _log_graph(self) -> None:
        """
        Saves the model graph.
        """
        x = self._get_filtered_input(self.dummy_input_train)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                self.log_writer.add_graph(self.net, x)
            except Exception:
                pass

    def _log_yaml(self) -> None:
        """
        Saves the yaml text on Tensorboard
        """
        # reading yaml
        with open(self.config_path) as f:
            txt = f.read()

        # adding also the parameters
        txt += f" parameters: {self._get_model_parameters() * 1e-3} K"

        self.log_writer.add_text("config.yaml", txt, 0)

    # = = = = = = = = = = = = = = = = = = = = = =
    #                Profiler
    def _get_profiler(self) -> torch.profiler.profile:
        """
        Return the profiler context manager.

        Returns
        -------
        torch.profiler.profile
            Pytorch profiler
        """
        prof = (
            torch.profiler.profile(
                schedule=torch.profiler.schedule(wait=1, warmup=1, active=3, repeat=2),
                on_trace_ready=torch.profiler.tensorboard_trace_handler(self.logs_dir),
                record_shapes=True,
                profile_memory=True,
                with_stack=True,
            )
            if self.enable_profiling
            else nullcontext()
        )
        return prof

    def _start_profiling(self) -> None:
        """
        Starts the profiling.
        """
        if self.enable_profiling:
            self.profiler.__enter__()
            logger.info("profiler started")

    def _stop_profiling(self) -> None:
        """
        Stops the profiling.
        """
        if self.enable_profiling:
            self.profiler.__exit__(None, None, None)
            logger.info("profiler stopped")

        # leaving

    # = = = = = = = = = = = = = = = = = = = = = =
    #                Utilities
    # = = = = = = = = = = = = = = = = = = = = = =
    def _remove_extra_dim(self, data: List[Tensor]) -> List[Tensor]:
        """
        Removes extra starting dimensions from the
        data loaded.

        Parameters
        ----------
        data : List[Tensor]
            List of Tensors returned from the Dataloader

        Returns
        -------
        List[Tensor]
            data without a possible extra leading dimension
        """
        data = [x[0] if x.shape[0] == 1 else x for x in data]
        return data

    def _get_model_parameters(self) -> int:
        """
        Gets the total model parameters.

        Returns
        -------
        int
            Number of net parameters
        """
        params = tu.get_model_parameters(self.net)
        return params

    def _get_filtered_input(self, data: List[Tensor]) -> List[Tensor]:
        """
        Filters the Dataloader output based on net_ins_indices.

        Parameters
        ----------
        data : List[Tensor]
            Dataloader output

        Returns
        -------
        List[Tensor]
            Filtered inputs for the network
        """
        return [data[i] for i in self.net_ins_indices]

    def _from_config(self, param: str, _type: Type, default: Any = None) -> Any:
        """
        gets a parameter from the training configuration.

        Parameters
        ----------
        param : str
            Name of the parameter
        _type : Type
            Parameter type
        default : Any, optional
            Default value, by default None

        Returns
        -------
        Any
            Value of the parameter
        """
        return self.config.get("training", param, _type, default)

    def _check_nan(self) -> None:
        """
        Checks if the last loss is NaN.

        Raises
        ----------
        RuntimeError
            In case a NaN value is encountered
        """
        if math.isnan(self.last_total_loss):
            err_msg = "NaN value encountered in last_total_loss"
            logger.error(err_msg)
            raise RuntimeError(err_msg)



from typing import Callable, List, Tuple, Type, Optional
from torch.utils.data import Dataset, DataLoader
from torch.optim import Adam, Rprop, Optimizer
from pathimport import set_module_root
from contextlib import nullcontext
from itertools import product
from torch import Tensor, nn
from loguru import logger
from pathlib import Path
import numpy as np
import unittest
import shutil
import torch

set_module_root("../torch_utils")
from tests.generate_test_data import get_test_data_dir
import torch_utilities as tu

# = = SETUP = = = = = = = = = = = = = = = = = = = = = = =


def _setup() -> None:
    np.random.seed(901)
    tu.set_device("auto")
    torch.manual_seed(984)
    torch.set_grad_enabled(True)
    logger.disable("torch_utilities.model_trainer")


# = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class ModelDummy(nn.Module):
    def __init__(self, config_path: Path) -> None:
        """
        Test model.
        """
        super().__init__()
        self.config = tu.Config(config_path)
        self.test_param = self.config.get("general", "test_param", int)
        self.layer = nn.Identity()
        self.batchnorm = nn.BatchNorm1d(8)

    def forward(self, x: Tensor) -> List[Tensor]:
        x = self.layer(x)
        x = self.batchnorm(x)
        return [x]


# = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class ModelTrainerDummy(tu.ModelTrainer):
    def __init__(
        self,
        model_path: Path,
        model: nn.Module,
        train_dl: DataLoader,
        valid_dl: DataLoader,
        optimizer_class: Type[Optimizer],
        losses: List[Callable],
        net_ins_indices: Optional[List[int]] = None,
        losses_names: Optional[List[str]] = None,
        enable_profiling: bool = False,
    ) -> None:
        """
        ModelTrainer implementation.
        """
        super().__init__(
            model_path,
            model,
            train_dl,
            valid_dl,
            optimizer_class,
            losses,
            net_ins_indices,
            losses_names,
            enable_profiling=enable_profiling,
        )
        self.net = self.net.to(float)

        # callbacks flags
        self.on_train_begin_flag = False
        self.on_train_end_flag = False
        self.on_train_step_begin_flag = False
        self.on_train_step_end_flag = False
        self.on_valid_step_begin_flag = False
        self.on_valid_step_end_flag = False

    def apply_losses(self, data: List[Tensor], net_outs: List[Tensor]) -> List[Tensor]:
        y_hat = net_outs[0]
        y_true = data[1]
        loss_values = [loss(y_hat, y_true) for loss in self.losses]
        return loss_values

    def tensorboard_logs(self, raw_data: List[Tensor], epoch: int) -> None:
        pass

    def on_train_begin(self) -> None:
        self.on_train_begin_flag = True

    def on_train_end(self) -> None:
        self.on_train_end_flag = True

    def on_train_step_begin(self, epoch) -> None:
        self.on_train_step_begin_flag = True

    def on_train_step_end(self, epoch) -> None:
        self.on_train_step_end_flag = True

    def on_valid_step_begin(self, epoch) -> None:
        self.on_valid_step_begin_flag = True

    def on_valid_step_end(self, epoch) -> None:
        self.on_valid_step_end_flag = True


# = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestModelTrainer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.model_path = [get_test_data_dir() / n for n in ("test_model", "test_model_overfit")]
        self.loader_type = (tu.HDF5Dataset, tu.HDF5OnlineDataset)
        self.optimizer = (Adam, Rprop)  # with and without weight decay setting
        self.n_losses = (4,)
        self.params = product(
            self.model_path,
            self.loader_type,
            self.optimizer,
            self.n_losses,
        )
        self.params = list(self.params)
        self.batch_size = 16
        self.clear_model_folders()

    def tearDown(self):
        self.clear_model_folders()

    def clear_model_folders(self):
        """
        Cleans the model folders.
        """
        for model in self.model_path:
            not_config = list(set(model.glob("*")) - set(model.glob("*.yml")))
            [shutil.rmtree(f) for f in not_config]

    def get_model_trainer(self, params: Tuple, enable_profiling: bool = False) -> ModelTrainerDummy:
        """
        Initializes the model trainer.

        Parameters
        ----------
        params : Tuple
            Parameters
        enable_profiling : bool, optional
            If True enables the profiler, by default False

        Returns
        -------
        ModelTrainerDummy
            ModelTrainer implementation
        """
        (
            model_path,
            loader_type,
            optimizer,
            n_losses,
        ) = params
        dataloader = self.get_dataloader(loader_type)
        trainer = ModelTrainerDummy(
            model_path=model_path,
            model=ModelDummy,
            train_dl=dataloader,
            valid_dl=dataloader,
            optimizer_class=optimizer,
            losses=[nn.MSELoss()] * n_losses,
            net_ins_indices=[0],
            losses_names=[f"test_loss_{i}" for i in range(n_losses)],
            enable_profiling=enable_profiling,
        )
        return trainer

    def get_dataloader(self, loader_type: Type[Dataset]) -> DataLoader:
        """
        Initialize a specific DataLoader.

        Parameters
        ----------
        loader_type : Type[Dataset]
            Dataset type

        Returns
        -------
        DataLoader
            DataLoader instance
        """
        dataset_path = get_test_data_dir() / "dataset.hdf5"
        data_layout = ["x", "y_true"]

        if loader_type == tu.HDF5Dataset:
            loader = tu.get_hdf5_dataloader(
                dataset_path=dataset_path,
                data_layout=data_layout,
                batch_size=self.batch_size,
            )
        else:
            n = 2
            ds = tu.HDF5OnlineDataset(
                dataset_paths=[dataset_path] * n,
                data_layouts=[data_layout] * n,
                batch_size=self.batch_size,
                total_items=self.batch_size,
            )
            loader = DataLoader(ds)

        return loader

    def test_init(self):
        for p in self.params:
            with self.subTest(p=p):
                self.get_model_trainer(p)

    def test_model_config(self):
        p = self.params[0]
        trainer = self.get_model_trainer(p)
        self.assertEqual(trainer.net.test_param, 1234)

    def test_learning_rate(self):
        p = self.params[0]
        trainer = self.get_model_trainer(p)
        optim = trainer.optimizer
        self.assertAlmostEqual(optim.param_groups[-1]["lr"], 0.001)

    def test_weight_decay(self):
        p = self.params[0]
        trainer = self.get_model_trainer(p)
        optim = trainer.optimizer
        self.assertAlmostEqual(optim.param_groups[-1]["weight_decay"], 0.002)

    def test_log_every(self):
        p = self.params[0]
        trainer = self.get_model_trainer(p)
        self.assertAlmostEqual(trainer.log_every, 1)

    def test_losses_weights(self):
        for p in self.params:
            with self.subTest(p=p):
                trainer = self.get_model_trainer(p)
                losses_w = trainer.losses_weights.tolist()
                target = list(range(1, len(losses_w) + 1))
                self.assertListEqual(losses_w, target)

    def test_callbacks(self):
        params = self.params[0], self.params[-1]
        for p in params:
            with self.subTest(p=p):
                overfit = "overfit" in str(p[0])
                trainer = self.get_model_trainer(p)
                trainer.start_training()
                self.assertEqual(
                    [
                        trainer.on_train_begin_flag,
                        trainer.on_train_end_flag,
                        trainer.on_train_step_begin_flag,
                        trainer.on_train_step_end_flag,
                        trainer.on_valid_step_begin_flag,
                        trainer.on_valid_step_end_flag,
                    ],
                    [True] * 4 + [False] * 2 if overfit else [True] * 6,
                )

    def test_profiler(self):
        p = self.params[0]
        for is_profiling in (False, True):
            with self.subTest(is_profiling=is_profiling):
                trainer = self.get_model_trainer(p, enable_profiling=is_profiling)
                expected = torch.profiler.profile if is_profiling else nullcontext
                self.assertEqual(type(trainer.profiler), expected)

    def test_get_dummy_data(self):
        params = self.params
        is_training = (False, True)
        params = product(params, is_training)
        for p, it in params:
            is_hdf5_dataset = p[1] == tu.HDF5Dataset
            with self.subTest(p=p, it=it):
                trainer = self.get_model_trainer(p)
                xs = trainer._get_dummy_data(it)
                y = 2 if is_hdf5_dataset else 4
                self.assertEqual(len(xs), y)
                y = 2 if is_hdf5_dataset else self.batch_size
                [self.assertEqual(x.shape[0], y) for x in xs]
                pass


if __name__ == "__main__":
    unittest.main(verbosity=2)



from pathimport import set_module_root
from loguru import logger
from pathlib import Path
import numpy as np
import yaml
import h5py

set_module_root(".")
from torch_utilities import save_audio


def create_data_dir():
    data_dir = Path(__file__).parent / "test_data"
    data_dir.mkdir(exist_ok=True)


def get_test_data_dir():
    path = Path(__file__).parent / "test_data"
    if not path.exists():
        raise FileNotFoundError("run generate_test_data.py first")
    return path


def generate_wavs(sample_rate: int = 16000, duration: float = 1):
    """
    Generate stereo and mono wav files
    of random noise.

    Parameters
    ----------
    sample_rate : int, optional
        Sample rate, by default 16000 Hz
    duration : float, optional
        Duration of the files, by default 1 s
    """
    data_dir = get_test_data_dir()
    files = [f"{x}.wav" for x in ("mono", "stereo")]
    channels = [1, 2]
    for n, c in zip(files, channels):
        x = np.stack([np.random.uniform(-1, 1, duration * sample_rate) for _ in range(c)])
        save_audio(data_dir / n, x, sample_rate)


def generate_hdf5(groups: int = 10, group_len: int = 16):
    data_layout = ["x", "y_true"]
    with h5py.File(get_test_data_dir() / "dataset.hdf5", "w") as ds:
        for i in range(groups):
            g = ds.create_group(f"group_{i}")
            for layout in data_layout:
                g_name = layout
                g.create_dataset(g_name, data=np.ones((group_len, 8)) * i)


def generate_yaml():
    """
    Generate a yaml dummy configuration.
    """
    data = dict(
        section1=dict(param1="test", param2=42, param3=["a", "b", "c"]),
        section2=dict(param4=12.43),
        ray_tune=dict(
            uniform=dict(loss_weight_0=[0, 1]),
            choice=dict(
                depth=[[1, 4, 10]],
                list_choice=[[[1, 2, 3], [0.1, 0.2, 0.3]]],
            ),
            loguniform=dict(
                learning_rate=[1e-5, 1e-3],
                weight_decay=[1e-2, 1e-1],
            ),
        ),
    )

    with open(get_test_data_dir() / "test.yml", "w") as f:
        yaml.dump(data, f)


def generate_test_model_config(name: str, overfit_mode: bool):
    txt = f"""
---
general:
  test_param: 1234
  
training:
  learning_rate: 0.001
  weight_decay: 0.002
  overfit_mode: {overfit_mode}
  max_epochs: 1
  batch_size: 16
  log_every: 1
  num_workeres: 1
  losses_weights:
    - 1 
    - 2 
    - 3 
    - 4 
    """
    tag = "_overfit" if overfit_mode else ""
    model_dir = get_test_data_dir() / f"test_model{tag}"
    model_dir.mkdir(exist_ok=True)
    config_path = model_dir / "config.yml"
    with open(config_path, "wt") as f:
        f.write(txt)


def generate_checkpoint_monitoring(filename: str, n: int) -> None:
    """
    Generates a dummy checkpoint monitoring list.

    Parameters
    ----------
    filename : str
        Filename
    n : int
        Number of checkpoints
    """
    _name = lambda i: f"checkpoint_{i}"
    checkpoints = [(_name(i), float(i)) for i in range(n)]
    with open(get_test_data_dir() / filename, "w") as f:
        yaml.dump(checkpoints, f)


def main():
    logger.info("generating test data")
    create_data_dir()
    generate_wavs()
    generate_hdf5()
    generate_yaml()
    generate_test_model_config("test_model", False)
    generate_test_model_config("test_model", True)
    generate_checkpoint_monitoring("dummy_ckpt.yml", 10)
    logger.info("generation complete")


if __name__ == "__main__":
    main()
